module Sem05 where


-- Сопоставление с образцом(паттернматчинг)
-- Пусть f определена как
-- f (~t : ~r : ~u : ~ "e") = (:) r $ (:) u [u]
-- Вернет ли f "False" что-нибудь? Почему? Если вернет, то что именно?

-- На каких образцах удачно сматичтся строка "Wow"?
-- (x : y) +
-- ((:) x y) +
-- [x, y] -
-- (x : y : z) +
-- (x : [y, z]) +
-- ([x, y, z]) +
-- (x : y : z : []) +
-- (x : y : z : w) +
-- (x : y : z : w : []) -
-- Какие значения будут связаны с переменными при удачном сопоставлении

-- Строгость и форсирование вычислений
-- Вообще говоря, seq ничего не гарантирует про порядок вычислений
-- Hoogle говорит, что выражение a `seq` b не гарантирует, что a вычислится перед b
-- Единственное, что гарантируется, это то, что оба выражения будут вычислены
-- перед тем, как seq вернет результат

-- Каков результат следующих выражений?
-- seq ((\True x -> "Blah") undefined) 123
-- seq ((\True -> \x -> "Blah") undefined) 123


-- Как вы думаете, почему?

-- Списки и функции над ними
-- В первой части запрещено использовать стандартные функции и генераторы.
-- Разрешена к использованию только рекурсия

-- Найдите количество четных элементов в данном списке
evenCount :: Integral a => [a] -> Integer
evenCount [] = 0
evenCount (x : xs)
    | even x = 1 + evenCount xs
    | otherwise = evenCount xs

-- По данному списку сформируйте список, содержащий только нечетные элементы исходного
onlyOdds :: Integral a => [a] -> [a]
onlyOdds [] = []
onlyOdds (x : xs)
    | odd x = x : onlyOdds xs
    | otherwise = onlyOdds xs

-- Дан список, сформируйте новый, в котором переставлены местами четные и нечетные(по порядку следования)
-- элементы исходного
-- [1, 2, 3, 4] => [2, 1, 4, 3]
-- [1, 2, 3] => [2, 1, 3]
permuteOddsAndEvens :: [a] -> [a]
permuteOddsAndEvens [] = []
permuteOddsAndEvens [x] = [x]
permuteOddsAndEvens (x : y : xs) = y : x : permuteOddsAndEvens xs

-- Даны два списка целых чисел. Сформируйте список, каждый элемент которого -- сумма соответствующих
-- элементов исходных списков. Обеспечьте разумное поведение для списков разной длины
pairwiseSum :: [Integer] -> [Integer] -> [Integer]
pairwiseSum [] _ = []
pairwiseSum _ [] = []
pairwiseSum (x : xs) (y : ys) = (x + y) : pairwiseSum xs ys


-- Разверните данный список. Постарайтесь обеспечить эффективную реализацию
reverse' :: [a] -> [a]
reverse' [] = []
reverse' (x : xs) = append x (reverse' xs)
    where
        append :: a -> [a] -> [a]
        append x' [] = [x']
        append x' (y:ys) = y : (append x' ys)

reverse'' :: [a] -> [a]
reverse'' xs = let
        rev_acc [] acc = acc
        rev_acc (x : xs') acc = rev_acc xs' (x : acc)
    in rev_acc xs []


-- Дан список и два числа k и n. Необходимо вытащить подсписок длины n, начиная c k-того по порядку
-- элемента исходного списка. Постарайтесь обеспечить разумное поведение для произвольных целых индексов
slice :: [a] -> Integer -> Integer -> [a]
slice [] _ _ = []
slice l@(x:xs) k n
    | n <= 0 = []
    | k < 0 = slice l 0 n
    | k == 0 = x : slice xs 0 (n - 1)
    | k > 0 = slice xs (k - 1) n


-- Во второй части запрещено использовать явную рекурсию.
-- Стандартные функции и генераторы разрешены к использованию

-- Напишите функцию, которая бы увеличивала все элементы списка в k раз
multiplyByK :: Num a => [a] -> a -> [a]
multiplyByK xs k = map (* k) xs

-- Напишите функцию, которая бы увеличивала элементы с четными значениями в k раз
multiplyEvensByK :: Integral a => [a] -> a -> [a]
multiplyEvensByK xs k = map (\x -> if even x then k * x else x) xs

-- Напишите функцию, которая по списку выдает список пар -- (индекс, элемент)
indices :: [a] -> [(Integer, a)]
indices = zip [0..]

pythrgs = [(x, y, z) | x <- [1..], y <- [1..x], z <- [1..(x + y)], (x^2) + (y^2) == (z^2)]

-- Напишите функцию, которая бы обнуляла элементы данного списка, неудовлетворяющие заданному условию
zeroBy :: Monoid a => [a] -> (a -> Bool) -> [a]
zeroBy xs p = map (\x -> if not $ p x then mempty else x) xs


-- Обобщите предыдущую функцию, написав функцию, которая бы применяла к элементам заданного списка,
-- неудовлетворяющим заданному условию заданную функцию
transformBy :: [a] -> (a -> Bool) -> (a -> a) -> [a]
transformBy xs p f = map (\x -> if not $ p x then f x else x) xs

-- Даны три списка целых чисел. Сформируйте список, каждый элемент которого -- сумма соответствующих
-- элементов исходных списков. Длина результата ограничена длиной самого короткого списка
triplewiseSum :: [Integer] -> [Integer] -> [Integer] -> [Integer]
triplewiseSum xs ys zs = zipWith3 (\x y z -> x + y + z) xs ys zs

data Foo = Foo {foo :: Integer, bar :: Double}
f = Foo {foo = 1, bar = 2.0}
